
using System;
using System.Collections.Generic;
using System.Linq;
using MonoTouch.Foundation;
using MonoTouch.UIKit;
using System.Net;
using System.Net.Sockets;
using Discovery.Core;
using System.Timers;

namespace ServiceLocationTest
{
	public partial class MainView : UIViewController
	{
		#region Constructors

		// The IntPtr and initWithCoder constructors are required for controllers that need 
		// to be able to be created from a xib rather than from managed code

		public MainView (IntPtr handle) : base(handle)
		{
			Initialize ();
		}

		[Export("initWithCoder:")]
		public MainView (NSCoder coder) : base(coder)
		{
			Initialize ();
		}

		public MainView () : base("MainView", null)
		{
			Initialize ();
		}

		void Initialize ()
		{
		}

		#endregion

		private List<string> _items = new List<string> ();
		public override void ViewDidLoad ()
		{
			base.ViewDidLoad ();
			
			table.DataSource = new TableViewDataSource (_items);
		}

		/// <summary>
		/// Called by our Ping button TouchUpInside event
		/// </summary>
		partial void Ping (UIButton sender)
		{
			// If we don't detect a wifi connection don't bother pinging
			if (Reachability.LocalWifiConnectionStatus () != NetworkStatus.ReachableViaWiFiNetwork) {
				var noWifiError = new UIAlertView ("WiFi Not Detected", "You do not appear to be connected to a WiFi network", null, "Ok, Sorry");
				noWifiError.Show ();
				return;
			}
			
			// Clear the server list
			ClearList ();
			
			// Setup listener and timeout
			StartListeningForServerPingBacks ();
			
			// Send out our location ping
			SendServerPing ();
		}

		#region Ping/Pong Code
		/// <summary>
		/// Simple state class that gets passed to our callback.
		/// 
		/// Used to get the state and start listening again.
		/// </summary>
		class UdpState
		{
			public readonly IPEndPoint endPoint;
			public readonly UdpClient client;

			public UdpState (IPEndPoint endPoint, UdpClient client)
			{
				this.endPoint = endPoint;
				this.client = client;
			}
		}

		/// <summary>
		/// Our callback that receives the pings back from the server(s)
		/// </summary>
		private void ReceiveServerPingCallback (IAsyncResult ar)
		{
			try {
				// Grab the state object and split it up
				UdpState state = (UdpState)(ar.AsyncState);
				UdpClient client = state.client;
				IPEndPoint endPoint = state.endPoint;
				
				// Grab all the data we received
				Byte[] receiveBytes = client.EndReceive (ar, ref endPoint);
				
				// Deserialize it to our ServiceInfo object
				var data = Discovery.Core.Helpers.DeserializeObject<ServiceInfo> (receiveBytes);
				
				// Using the UI thread, update the server list
				InvokeOnMainThread (delegate { AddItemToList (String.Format ("Name: {0} Endpoint: {1}", data.ServiceName, data.EndpointAddress)); });
				
				// Start listening again
				client.BeginReceive (new AsyncCallback (this.ReceiveServerPingCallback), state);
			} catch (Exception) {
				// Just in case we have any network issues, or we've closed the socket, we catch everything.
				// Rather a horrible catch all than an app crash in this instance.
			}
		}

		/// <summary>
		/// Initiate listening for server pings and set a timeout
		/// </summary>
		private void StartListeningForServerPingBacks ()
		{
			// Disable our ping button
			PingButton.Enabled = false;
			
			// Listen on our IP addresses on our port
			// In a real scenario you'd probably want to make sure the port
			// was available and fallback to an alternative if not.
			var endPoint = new IPEndPoint (GetCurrentIPAddress(), 8762);
			
			// Make sure we don't grab exclusive "rights" to our address
			// so we can use the same port for send and receive.
			var udpClient = new UdpClient ();
			udpClient.Client.SetSocketOption (SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
			udpClient.ExclusiveAddressUse = false;
			udpClient.Client.Bind (endPoint);
			
			// Setup our "state" object so the callback has access to the client and endpoint
			UdpState state = new UdpState (endPoint, udpClient);
			
			// Setup our async receive
			// Our callback will be called if and when data comes in
			udpClient.BeginReceive (new AsyncCallback (this.ReceiveServerPingCallback), state);
			
			// Setup a timeout timer.
			// When the timer elapses we enable our ping button again and
			// close our udpclient.
			var enableTimer = new Timer (2000);
			enableTimer.AutoReset = false;
			enableTimer.Elapsed += delegate(object sender, ElapsedEventArgs e) {
				InvokeOnMainThread (delegate { PingButton.Enabled = true; });
				udpClient.Close ();
			};
			enableTimer.Enabled = true;
		}

		/// <summary>
		/// Send our our server ping
		/// </summary>
		private void SendServerPing ()
		{
			// Create a socket udp ipv4 socket
			using (Socket sock = new Socket (AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp)) {
				// Create our endpoint using the IP broadcast address and our port
				IPEndPoint endPoint = new IPEndPoint (IPAddress.Broadcast, Discovery.Core.ServiceInfo.UDPPort);
				
				// Serialize our ping "payload"
				byte[] data = Discovery.Core.Helpers.SerializeObject<ServiceClientInfo> (new ServiceClientInfo (GetCurrentIPAddress ().ToString (), 8762));
				
				// Tell our socket to reuse the address so we can send and
				// receive on the same port.
				sock.SetSocketOption (SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
				
				// Tell the socket we want to broadcast - if we don't do this it
				// won't let us send.
				sock.SetSocketOption (SocketOptionLevel.Socket, SocketOptionName.Broadcast, true);
				
				// Send the ping and close the socket.
				sock.SendTo (data, endPoint);
				sock.Close ();
			}
		}
		
		/// <summary>
		/// Attempts to get the current "public" IPv4 address
		/// </summary>
		IPAddress GetCurrentIPAddress ()
		{
			IPAddress[] localIPs = null;
			
			// Try with GetHostName, add .local if that fails
			// We need to do this as there seems to be a bug / inconsistency
			// between the simulator and the device itself
			try 
			{
				localIPs = Dns.GetHostAddresses (Dns.GetHostName ());
			} 
			catch (SocketException) 
			{
			}
			
			if (localIPs == null) {
				try 
				{
					localIPs = Dns.GetHostAddresses (Dns.GetHostName () + ".local");
				} 
				catch (SocketException) 
				{
					// fallback to loopback
					return IPAddress.Loopback;
				}
			}
			
			// Work through the IPs reported and return the first one
			// that is IPv4 and not a loopback
			foreach (IPAddress localIP in localIPs) {
				if ((localIP.AddressFamily == System.Net.Sockets.AddressFamily.InterNetwork) && (!IPAddress.IsLoopback (localIP)))
					return localIP;
			}
			
			// Fallback to loopback if necessary
			return IPAddress.Loopback;
		}

		#endregion

		#region Server List
		private void AddItemToList (string item)
		{
			_items.Add (item);
			table.ReloadData ();
		}

		private void ClearList ()
		{
			_items.Clear ();
			table.ReloadData ();
		}

		private class TableViewDataSource : UITableViewDataSource
		{
			public List<string> List { get; private set; }
			public TableViewDataSource (List<string> list)
			{
				List = list;
			}
			public override int RowsInSection (UITableView tableview, int section)
			{
				return List.Count ();
			}
			public override UITableViewCell GetCell (UITableView tableView, NSIndexPath indexPath)
			{
				var cell = tableView.DequeueReusableCell ("Me");
				
				if (cell == null)
					cell = new UITableViewCell (UITableViewCellStyle.Default, "Me");
				
				cell.TextLabel.Text = List[indexPath.Row];
				
				return cell;
			}
			
			
		}
		#endregion
		
	}
	
	
}
